//---------------------------------------------------------------------------
//
//   MoveNormal
//
//          Copyright(C) 2013, Tark
//
//  @Ɉړ
//
//
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <unordered_map>
#include <unordered_set>
#include <algorithm>
#include "MQPlugin.h"
#if MQPLUGIN_VERSION >= 0x0400
#include "MQ3DLib.h"
#include "MQWidget.h"
#endif
#include "CMQObject.h"
#include "MoveNormal.h"

static MoveNormal plugin;

//---------------------------------------------------------------------------
//  MQGetPlugInID
//---------------------------------------------------------------------------
void MoveNormal::GetPlugInID(DWORD *Product, DWORD *ID)
{
	*Product = 0xE3F7A830;
	*ID      = 0xBD9A186A;
}

//---------------------------------------------------------------------------
//  MQGetPlugInName
//---------------------------------------------------------------------------
const char *MoveNormal::GetPlugInName(void)
{
	return "MoveNormal ver.1.0 Copyright(C) 2013, Tark.";
}

//---------------------------------------------------------------------------
//  MQEnumString
//---------------------------------------------------------------------------
const char *MoveNormal::EnumString()
{
	return "MoveNormal";
}

//---------------------------------------------------------------------------
//  Initialize
//---------------------------------------------------------------------------
BOOL MoveNormal::Initialize()
{
	if (m_Window == NULL){
		MQWindowBase mainwnd = MQWindow::GetMainWindow();
		m_Window = new MoveNormalWindow(this, mainwnd);
	}

	return TRUE;
}

//---------------------------------------------------------------------------
//  Exit
//---------------------------------------------------------------------------
void MoveNormal::Exit()
{
	if (m_Window != NULL){
		delete m_Window;
		m_Window = NULL;
	}
}

//---------------------------------------------------------------------------
//  Activate
//---------------------------------------------------------------------------
BOOL MoveNormal::Activate(MQDocument doc, BOOL flag)
{
	if (!m_Window) {
		return FALSE;
	}
	m_bActivate = flag ? true : false;
	m_Window->SetVisible(m_bActivate);
	if(m_bActivate) {
		m_normal_update = true;
		MQ_RefreshView(nullptr);
	}
	else {
		DeleteDrawingObject(m_doc, m_draw);
		m_draw = nullptr;
		m_obj = nullptr;
		m_doc = nullptr;
	}
	return m_bActivate;
}

//---------------------------------------------------------------------------
//  IsActivated
//---------------------------------------------------------------------------
BOOL MoveNormal::IsActivated(MQDocument doc)
{
	if (m_Window == NULL) {
		return FALSE;
	}
	return m_bActivate;
}

// vOC̃x[XNXԂ
MQBasePlugin *GetPluginClass()
{
	return &plugin;
}

MoveNormal::MoveNormal() :
		m_Window(nullptr), m_bActivate(false),
		m_val(5.0f), m_doc(nullptr), m_obj(nullptr), m_draw(nullptr),
		m_normal_update(true), m_draw_update(true),
		m_viewLine(true), m_checkNeighbor(true)
{
}

MoveNormal::~MoveNormal()
{
	if(m_draw) {
		DeleteDrawingObject(m_doc, m_draw);
		m_draw = nullptr;
	}
	if(!m_Window){
		delete m_Window;
		m_Window = nullptr;
	}
}

void MoveNormal::createNormalViewObject()
{
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();

	m_normal.clear();
	// _̗אږ
	std::unordered_multimap<cmq::vindex_t, cmq::findex_t> vfmap;
	// אڑI/ӂ钸_
	std::unordered_multimap<cmq::vindex_t, cmq::findex_t> vfmapselect;
	cmq::CMQContainer<cmq::CMQFace, cmq::CMQObject> fcontainer(&obj);
	for(auto fit = fcontainer.begin(); fit != fcontainer.end(); ++fit) {
		std::vector<int> varray = fit->getFacePointArray();
		for(int i = 0; i < varray.size(); i++) {
			cmq::vindex_t vindex = varray[i];
			int vcount = varray.size();
			// אږ/IĂ
			if(fit->isSelect() ||
				doc.isSelectLine(obj, fit.index(), i) |
				doc.isSelectLine(obj, fit.index(), (i + vcount - 1) % vcount)) {
				vfmapselect.insert(std::make_pair(vindex, fit.index()));
			}
			vfmap.insert(std::make_pair(vindex, fit.index()));
		}
	}
	// ʂ̖@
//	std::unordered_map<cmq::findex_t, MQPoint> fnormal;
	cmq::CMQContainer<cmq::CMQVertex, cmq::CMQObject> vcontainer(&obj);
	for(auto vit = vcontainer.begin(); vit != vcontainer.end(); ++vit) {
		MQPoint normal(0, 0, 0);
		int normalcount = 0;
		if(!vit->isSelect()) {
			auto selectedface = vfmapselect.equal_range(vit.index());
			if(selectedface.first == selectedface.second) {
				// I𒸓_אڑI/ӂȂ͏ΏۊO
				continue;
			}
		}
		auto found = vfmap.equal_range(vit.index());
		auto selectedface = vfmapselect.equal_range(vit.index());
		for(auto it = found.first; it != found.second; ++it) {
			// eאږ/ӂɂ
			auto find = std::find_if(selectedface.first, selectedface.second, [&](const std::pair<cmq::vindex_t, cmq::findex_t> &p) {
				return p.second == it->second;
			});
			if(find == selectedface.second) {
				// I
				if(!vit->isSelect()) {
					// I𒸓_
					continue;
				}
				else if(!m_checkNeighbor && selectedface.first != selectedface.second){
					// I𒸓_ŗאږʎQOFFɗאڑIʂ
					continue;
				}
			}
			// @vZ
			auto varray = obj.getFacePointArray(it->second);
			auto findit = std::find(varray.begin(), varray.end(), vit.index());
			if(findit == varray.end()) {
				//something wrong
				continue;
			}
			MQPoint nvec = obj.getNormalVector(it->second, *findit);
			nvec.normalize();
			normal += nvec;
			normalcount++;
		}

		if(normalcount > 0) {
			normal /= (float)normalcount;
			m_normal.insert(std::make_pair(vit.index(), normal));
		}
	}
}

void MoveNormal::OnDraw(MQDocument doc_, MQScene scene, int width, int height)
{
	if(!IsActivated(doc_)) {
		return;
	}
	m_doc = doc_;
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();

	if(m_normal_update) {
		createNormalViewObject();
		m_normal_update = false;
		m_draw_update = true;
		DeleteDrawingObject(m_doc, m_draw);
		m_draw = nullptr;
	}

	if(!m_viewLine) {
		if(m_draw) {
			DeleteDrawingObject(m_doc, m_draw);
		}
	}
	else if(m_draw_update) {
		if(m_draw) {
			DeleteDrawingObject(m_doc, m_draw);
		}
		m_draw = CreateDrawingObject(m_doc, DRAW_OBJECT_LINE, FALSE);
		m_draw->SetColor(MQColor(1,1,1));
		m_draw->SetColorValid(TRUE);
		cmq::CMQObject draw(m_draw, &doc);
		for(auto found : m_normal) {
			int vidx = found.first;
			MQPoint normal = found.second;
			MQPoint v1 = obj.getVertex(vidx);
			MQPoint v2(v1 + normal * m_val);
			draw.createLine(v1, v2);
		}
		m_draw_update = false;
	}
}

BOOL MoveNormal::OnLeftButtonDown(MQDocument doc_, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!IsActivated(doc_)) {
		return FALSE;
	}
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();
	m_vertex_premoved.clear();
	for(auto n : m_normal) {
		int vidx = n.first;
		m_vertex_premoved.insert(std::make_pair(vidx, obj.getVertex(vidx)));
	}
	m_MousePos_x = state.MousePos.x;
	return TRUE;
}

BOOL MoveNormal::OnLeftButtonMove(MQDocument doc_, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!IsActivated(doc_)) {
		return FALSE;
	}
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();

	float moved = (float)(state.MousePos.x - m_MousePos_x);
	if(state.Shift) {
		moved *= 5.0f;
	}
	else if(state.Ctrl) {
		moved /= 20.0f;
	}
	else {
		moved /= 2.0;
	}

	for(auto n : m_normal) {
		int vidx = n.first;
		auto found = m_vertex_premoved.find(vidx);
		if(found != m_vertex_premoved.end()) {
			MQPoint origpos = found->second;
			MQPoint normal = n.second;
			MQPoint newpos = origpos + normal * moved;
			obj.setVertex(vidx, newpos);
		}
	}
	m_draw_update = true;
	MQ_RefreshView(nullptr);
	return TRUE;
}

BOOL MoveNormal::OnLeftButtonUp(MQDocument doc_, MQScene scene, MOUSE_BUTTON_STATE& state)
{
	if(!IsActivated(doc_)) {
		return FALSE;
	}
	UpdateUndo();
	return TRUE;
}
BOOL MoveNormal::OnUndo(MQDocument doc, int undo_state)
{
	if(!IsActivated(doc)) {
		return FALSE;
	}
	m_draw_update = true;
	return FALSE;
}

void MoveNormal::OnUpdateObjectList(MQDocument doc)
{
	if(IsActivated(doc)) {
		if(doc->GetCurrentObjectIndex() != doc->GetObjectIndex(m_obj)) {
			m_normal_update = true;
		}
	}
}

BOOL MoveNormal::move(float val)
{
	cmq::CMQDocument doc(m_doc);
	cmq::CMQObject obj = doc.getCurrentObject();
	for(auto n : m_normal) {
		int vidx = n.first;
		MQPoint normal = n.second;
		MQPoint pos = obj.getVertex(vidx);
		pos += normal * val;
		obj.setVertex(vidx, pos);
	}
	m_draw_update = true;
	UpdateUndo();
	MQ_RefreshView(nullptr);
	return TRUE;
}

void MoveNormal::valChange(float val)
{
	m_val = val;
	m_draw_update = true;
}

void MoveNormal::viewLine(bool flag)
{
	m_viewLine = flag;
	m_draw_update = true;
}

void MoveNormal::checkNeighbor(bool flag)
{
	m_checkNeighbor = flag;
	m_normal_update = true;
}

#if MQPLUGIN_VERSION >= 0x0400

MoveNormalWindow::MoveNormalWindow(MoveNormal *plugin, MQWindowBase& parent) : MQWindow(parent), m_pPlugin(plugin)
{
	setlocale(LC_CTYPE, "");
	wchar_t buffer[256] = L"MoveNormal";
	SetTitle(buffer);

	MQFrame *mainFrame = CreateVerticalFrame(this);
	spin = CreateDoubleSpinBox(mainFrame);
	spin->SetPosition(5.0);
	spin->SetMin(-100.0);
	spin->SetMax(100.0);
	spin->SetIncrement(1.0);
	spin->SetMaxDecimalDigit(3);
	spin->AddChangingEvent(this, &MoveNormalWindow::valChange);

	MQFrame *frame1 = CreateHorizontalFrame(mainFrame);
	MQButton *buttonPlus = CreateButton(frame1, L"+");
	buttonPlus->AddClickEvent(this, &MoveNormalWindow::clickPlus);
	MQButton *buttonMinus = CreateButton(frame1, L"-");
	buttonMinus->AddClickEvent(this, &MoveNormalWindow::clickMinus);

	MQFrame *frame2 = CreateVerticalFrame(mainFrame);
	checkViewLine = CreateCheckBox(frame2, L"View Line(&V)");
	checkViewLine->AddChangedEvent(this, &MoveNormalWindow::viewLineChanged);
	checkViewLine->SetChecked(true);
	checkCheckNeighbor = CreateCheckBox(frame2, L"Check Neighbor(&N)");
	checkCheckNeighbor->AddChangedEvent(this, &MoveNormalWindow::checkNeighborChanged);
	checkCheckNeighbor->SetChecked(true);

	this->AddKeyUpEvent(this, &MoveNormalWindow::keyevent);
}

MoveNormalWindow::~MoveNormalWindow()
{
}

BOOL MoveNormalWindow::clickPlus(MQWidgetBase *sender, MQDocument doc)
{
	float val = spin->GetPosition();
	return m_pPlugin->move(val);
}

BOOL MoveNormalWindow::clickMinus(MQWidgetBase *sender, MQDocument doc)
{
	float val = spin->GetPosition();
	return m_pPlugin->move(-val);
}

BOOL MoveNormalWindow::keyevent(MQWidgetBase *sender, MQDocument doc, MQWidgetKeyParam& param)
{
	if(param.Alt) {
		switch(param.Key) {
		case 0x56: checkViewLine->SetChecked(!checkViewLine->GetChecked()); break;
		case 0x4E: checkCheckNeighbor->SetChecked(!checkCheckNeighbor->GetChecked()); break;
		}
	}
	return FALSE;
}

BOOL MoveNormalWindow::valChange(MQWidgetBase *parent, MQDocument doc)
{
	float val = spin->GetPosition();
	m_pPlugin->valChange(val);
	return TRUE;
}

BOOL MoveNormalWindow::viewLineChanged(MQWidgetBase *parent, MQDocument doc)
{
	bool flag = checkViewLine->GetChecked();
	m_pPlugin->viewLine(flag);
	return TRUE;
}

BOOL MoveNormalWindow::checkNeighborChanged(MQWidgetBase *parent, MQDocument doc)
{
	bool flag = checkCheckNeighbor->GetChecked();
	m_pPlugin->checkNeighbor(flag);
	return TRUE;
}

#else

#endif
